from typing import TypeVar
from .vmath import Vector2d, Angle
from pygame import Surface

class Component:
    game_object: "GameObject"

    class ComponentData:
        component_type: type

        def __init__(self, ctp: type):
            self.component_type = ctp

    def draw(self, display: Surface):
        pass

    def update(self):
        pass

    def destroy(self):
        pass

    def get_data(self) -> ComponentData:
        return self.ComponentData(Component)


class Transform:
    game_object: "GameObject"
    position: Vector2d  # position of CENTER
    angle: Angle
    scale: float
    childs: list["Transform"]

    def __init__(self, obj: "GameObject"):
        self.game_object = obj
        self.position = Vector2d(0, 0)
        self.angle = Angle()
        self.scale = 1
        self.childs = []

    def translate(self, trn: Vector2d):
        self.position += trn
        for child in self.childs:
            child.translate(trn)

    def rotate(self, angle: Angle):
        self.angle += angle
        # TODO rotate all of the childs relative to center of this transform.

    def add_child(self, child: "Transform"):
        self.childs.append(child)


T = TypeVar("T", bound=Component)


class GameObject:
    """
    Inspired by Unity.
    """
    components: list[type[Component]]
    active: bool
    tag: str
    transform: Transform
    childs: list["GameObject"]

    objects: list["GameObject"] = []

    def __init__(self, tag: str):
        self.components = []
        self.active = True
        self.tag = tag
        self.transform = Transform(self)
        GameObject.objects.append(self)

    def draw(self, display: Surface):
        for component in self.components:
            component.draw(display)
        if not "child" in self.__dir__():
            return 
        for child in self.childs:
            child.draw(display)

    def update(self):
        for i in self.components:
            i.update()

    def on_destroy(self):
        for i in self.childs:
            GameObject.destroy(i)

    def add_component(self, component: Component):
        self.components.append(component)
        component.game_object = self

    def get_component(self, component: type[T]) -> T:
        for i in self.components:
            if isinstance(i, component):
                return i
        raise Exception(f"No such component: {component}")

    def set_active(self, active: bool):
        self.active = active

    def clone(self) -> "GameObject":
        # TODO
        raise NotImplemented()

    @staticmethod
    def get_by_tag(tag: str) -> "GameObject":
        for i in GameObject.objects:
            if i.tag == tag:
                return i
        raise Exception("No such object")

    @staticmethod
    def destroy(obj: "GameObject"):
        obj.on_destroy()
        del obj


class Prefab:
    """
    Saves all the data about the object. Can be saved or loaded from the file.
    """
    components: list[Component.ComponentData]

    def __init__(self, game_object: GameObject):
        self.components = []
        for i in game_object.components:
            self.components.append(i.get_data())

    def load(self, filename: str):
        raise NotImplemented

    def save(self):
        raise NotImplemented


class GameObjectData:
    """
    Shorthand way to register a prefab to scene.
    """
    position: Vector2d
    angle: Angle
    components: list[Component]

    def __init__(self, pos=Vector2d(), angle=Angle(), *args: Component):
        self.position = pos
        self.angle = angle
        self.components = list(args)





from math import pi, sqrt
from typing import Callable


class Vector2d:
    """
    Class to represent a pair of floats.
    """

    x: float
    y: float

    def __init__(self, a: float = 0, b: float = 0):
        self.x = a
        self.y = b

    @staticmethod
    def from_tuple(tpl: tuple[float, float]) -> "Vector2d":
        return Vector2d(tpl[0], tpl[1])

    def as_tuple(self) -> tuple[float, float]:
        return self.x, self.y

    def distance(self, other: "Vector2d") -> float:
        return sqrt(((self.x - other.x) ** 2) + ((self.y - other.y) ** 2))

    def intx(self) -> int:
        return int(self.x)

    def inty(self) -> int:
        return int(self.y)

    def distanceLooped(self, other: "Vector2d", size: "Vector2d") -> float:
        return sqrt(((self.x - other.x) ** 2) % ((size.x // 2) ** 2) + ((self.y - other.y) ** 2) % ((size.y // 2) ** 2))

    def __add__(self, other: "Vector2d") -> "Vector2d":
        return Vector2d(self.x + other.x, self.y + other.y)

    def __sub__(self, other: "Vector2d") -> "Vector2d":
        return Vector2d(self.x - other.x, self.y - other.y)

    def __mul__(self, other: float):
        return Vector2d(self.x * other, self.y * other)
    
    def complexMultiply(self, other: "Vector2d") -> "Vector2d":
        # complex multiplying
        return Vector2d(self.x * other.x - self.y * other.y, self.y * other.x + self.x * other.y)

    def __truediv__(self, other: float):
        return Vector2d(self.x / other, self.y / other)

    def operation(self, other: "Vector2d", operation: Callable[[float, float], float]) -> "Vector2d":
        return Vector2d(operation(self.x, other.x), operation(self.y, other.y))

    def __repr__(self) -> str:  # for debugging
        return "<" + str(self.x) + ", " + str(self.y) + ">"

    def __eq__(self, other: "Vector2d") -> bool:
        return (self.x == other.x and self.y == other.y)

    def __ne__(self, other: "Vector2d") -> bool:
        return (self.x != other.x or self.y != other.y)


class Angle:
    """
    class that represent angles in radians
    """

    angle: float

    def __init__(self, angle: float = 0) -> None:
        self.angle = angle
        self.bound()

    def set(self, angle: float, isDegree: bool = False):
        if isDegree:
            angle = angle * pi / 180
        self.angle = angle
        self.bound()

    def get(self, isDegree: bool = False):
        if isDegree:
            return self.angle * 180 / pi
        return self.angle

    def bound(self):
        self.angle %= (2 * pi)

    def __add__(self, other: "Angle") -> "Angle":
        return Angle(self.get() + other.get())

    def __sub__(self, other: "Angle") -> "Angle":
        return Angle(self.get() - other.get())

    def __repr__(self) -> str:
        return str(self.angle)

    def __float__(self) -> float:
        return self.angle

class Mod4:
    def __init__(self, value: int = 0) -> None:
        self.value = value % 4
    
    def __add__(self, other: "Mod4") -> "Mod4":
        return Mod4((self.value + other.value) % 4)
    
    def __sub__(self, other: "Mod4") -> "Mod4":
        return Mod4((self.value - other.value) % 4)

    def __repr__(self) -> str:
        return self.value.__repr__()

    def __eq__(self, other: "Mod4") -> bool:
        return self.value == other.value

    def __ne__(self, other: "Mod4") -> bool:
        return self.value != other.value

class Direction(Mod4):
    def __init__(self, value: int = 0) -> None:
        super().__init__(value)
    
    def toAngle(self) -> Angle:
        return Angle(self.value * pi / 2)

    @staticmethod
    def fromAngle(ang: Angle) -> "Direction":
        return Direction((ang.get() + pi/4) // (pi/2))
    
    def toVector2d(self) -> Vector2d:
        return Directions.AsVector2D[self.value]

    @staticmethod
    def fromVector2d(ang: Vector2d) -> "Direction":
        newang = ang.complexMultiply(Vector2d(1, 1))
        return Direction(int(newang.y < 0) * 2 + int(newang.x < 0))

    def rotateVector2d(self, vec: Vector2d) -> Vector2d:
        return vec.complexMultiply(self.toVector2d())
    
class Directions:
    RIGHT = Direction(0)
    UP = Direction(1)
    LEFT = Direction(2)
    DOWN = Direction(3)
    
    AsVector2D = (
        Vector2d(1, 0), 
        Vector2d(0, -1), 
        Vector2d(-1, 0), 
        Vector2d(0, 1)
    )



from .widgets import *

widgets: dict[str, UI] = dict()


def add_menu(nm: str):
    widgets[nm] = UI(nm)
    return widgets[nm]


def draw_all(window: pygame.Surface):
    for v in widgets.values():
        v.draw(window)


def update_all(events: list[pygame.event.Event]):
    for v in widgets.values():
        v.update(events)


def get_menu(nm: str):
    return widgets[nm]



import pygame

from ..game_object import Component
from pygame import Surface

class SpriteComponent(Component):

    def __init__(self, sprite_name: str, size: tuple[int, int]) -> None:
        super().__init__()

        self.size = size
        self.texture = pygame.image.load("./sprites/" + sprite_name).convert_alpha()

    def draw(self, display: Surface):
        blitImage = self.texture

        cropped = pygame.Surface(self.size)
        cropped.blit(blitImage, (0, 0))

        angle = self.game_object.transform.angle.get()
        scale = self.game_object.transform.scale

        if angle != 0:
            cropped = pygame.transform.rotate(cropped, angle)

        if scale != 1:
            cropped = pygame.transform.scale_by(cropped, scale)

        rect = cropped.get_rect(center=self.game_object.transform.position.as_tuple())

        pygame.draw.rect(display, (255, 0, 0), rect, 1)
        display.blit(blitImage, rect)
    


from .vmath import *


class Camera:
    x: float
    y: float
    zoom: float

    def __init__(self) -> None:
        self.x = 0
        self.y = 0
        self.zoom = 0

    def _normalize(self, pos: Vector2d) -> Vector2d:
        """
        Normalizes pos coordinates to relative to camera coordinates.
        """

        # TODO: Zoom in/out
        return Vector2d(pos.x - self.x, pos.y - self.y)

    def normalize(self, pos: Vector2d | tuple[float, float]) -> Vector2d:
        if isinstance(pos, tuple):
            return self._normalize(Vector2d.from_tuple(pos))
        return self._normalize(pos)

    def set_x(self, x: float):
        self.x = x

    def set_y(self, y: float):
        self.y = y

    def set_position(self, pos: Vector2d | tuple[float, float]):
        if isinstance(pos, tuple):
            pos = Vector2d.from_tuple(pos)
        self.x = pos.x
        self.y = pos.y

    def add_x(self, x: float):
        self.x += x

    def add_y(self, y: float):
        self.y += y

    def add_pos(self, pos: Vector2d | tuple[float, float]):
        if isinstance(pos, tuple):
            pos = Vector2d.from_tuple(pos)
        self.x += pos.x
        self.y += pos.y

    def get_x(self):
        return self.x

    def get_y(self):
        return self.y

    def get_pos(self):
        return Vector2d(self.x, self.y)


camera = Camera()



# Use only if everything is bad.
class CriticalError(Exception):
    message: str | None

    def __init__(self, message: str | None = None):
        self.message = message

    def __str__(self):
        if self.message:
            return f"Attention, CriticalError! {self.message}"
        return f"Attention, CriticalError!"


# May be useless(?)
class WarningError(Exception):
    message: str | None
    cause: Exception | None

    def __init__(self, message: str | None = None, cause: Exception | None = None):
        self.cause = cause
        self.message = message

    def __str__(self):
        s = "Warning!"
        if self.message: s += f" Message: {self.message}"
        if self.cause: s += f", Cause: {self.cause}"
        return s


class NoComponentError(Exception):

    def __init__(self, component):
        pass
